# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Specification for 'painted' items. See https://github.com/RLBot/RLBot/wiki/Bot-Customization
class LoadoutPaint(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsLoadoutPaint(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = LoadoutPaint()
        x.Init(buf, n + offset)
        return x

    # LoadoutPaint
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # LoadoutPaint
    def CarPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # LoadoutPaint
    def DecalPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # LoadoutPaint
    def WheelsPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # LoadoutPaint
    def BoostPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # LoadoutPaint
    def AntennaPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # LoadoutPaint
    def HatPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # LoadoutPaint
    def TrailsPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # LoadoutPaint
    def GoalExplosionPaintId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def LoadoutPaintStart(builder): builder.StartObject(8)
def LoadoutPaintAddCarPaintId(builder, carPaintId): builder.PrependInt32Slot(0, carPaintId, 0)
def LoadoutPaintAddDecalPaintId(builder, decalPaintId): builder.PrependInt32Slot(1, decalPaintId, 0)
def LoadoutPaintAddWheelsPaintId(builder, wheelsPaintId): builder.PrependInt32Slot(2, wheelsPaintId, 0)
def LoadoutPaintAddBoostPaintId(builder, boostPaintId): builder.PrependInt32Slot(3, boostPaintId, 0)
def LoadoutPaintAddAntennaPaintId(builder, antennaPaintId): builder.PrependInt32Slot(4, antennaPaintId, 0)
def LoadoutPaintAddHatPaintId(builder, hatPaintId): builder.PrependInt32Slot(5, hatPaintId, 0)
def LoadoutPaintAddTrailsPaintId(builder, trailsPaintId): builder.PrependInt32Slot(6, trailsPaintId, 0)
def LoadoutPaintAddGoalExplosionPaintId(builder, goalExplosionPaintId): builder.PrependInt32Slot(7, goalExplosionPaintId, 0)
def LoadoutPaintEnd(builder): return builder.EndObject()



from multiprocessing import Event
from signal import SIGTERM
from subprocess import Popen

import rlbot
from rlbot.agents.base_agent import BOT_CONFIG_AGENT_HEADER
from rlbot.agents.base_independent_agent import BaseIndependentAgent
from rlbot.parsing.custom_config import ConfigHeader, ConfigObject
from rlbot.utils.structures.game_interface import get_dll_directory


class BaseSubprocessAgent(BaseIndependentAgent):
    """An Agent which simply delegates to an external process."""

    path: str

    @staticmethod
    def create_agent_configurations(config: ConfigObject) -> None:
        params = config.get_header(BOT_CONFIG_AGENT_HEADER)
        params.add_value('path', str, description='The path to the exe for the bot')

    def load_config(self, config_header: ConfigHeader) -> None:
        self.path = config_header.getpath('path')

    def run_independently(self, terminate_request_event: Event) -> None:
        # This argument sequence is consumed by Rust's `rlbot::run`.
        process = Popen([
            self.path,
            '--rlbot-version',
            rlbot.__version__,
            '--rlbot-dll-directory',
            get_dll_directory(),
            '--player-index',
            str(self.index),
        ])

        # Block until we are asked to terminate, and then do so.
        terminate_request_event.wait()
        process.send_signal(SIGTERM)
        process.wait()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class RenderMessage(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRenderMessage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RenderMessage()
        x.Init(buf, n + offset)
        return x

    # RenderMessage
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RenderMessage
    def RenderType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 1

    # RenderMessage
    def Color(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Color import Color
            obj = Color()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// For 2d renders this only grabs x and y
    # RenderMessage
    def Start(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// For 2d renders this only grabs x and y
    # RenderMessage
    def End(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// Scales the x size of the text/rectangle, is used for rectangles assuming an initial value of 1
    # RenderMessage
    def ScaleX(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 1

# /// Scales the y size of the text/rectangle, is used for rectangles assuming an initial value of 1
    # RenderMessage
    def ScaleY(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 1

    # RenderMessage
    def Text(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return bytes()

# /// Rectangles can be filled or just outlines.
    # RenderMessage
    def IsFilled(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

def RenderMessageStart(builder): builder.StartObject(8)
def RenderMessageAddRenderType(builder, renderType): builder.PrependInt8Slot(0, renderType, 1)
def RenderMessageAddColor(builder, color): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(color), 0)
def RenderMessageAddStart(builder, start): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(start), 0)
def RenderMessageAddEnd(builder, end): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(end), 0)
def RenderMessageAddScaleX(builder, scaleX): builder.PrependInt32Slot(4, scaleX, 1)
def RenderMessageAddScaleY(builder, scaleY): builder.PrependInt32Slot(5, scaleY, 1)
def RenderMessageAddText(builder, text): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(text), 0)
def RenderMessageAddIsFilled(builder, isFilled): builder.PrependBoolSlot(7, isFilled, 0)
def RenderMessageEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class ExistingMatchBehavior(object):
# /// Restart the match if any match settings differ. This is the default because old RLBot always worked this way.
    Restart_If_Different = 0
# /// Always restart the match, even if config is identical
    Restart = 1
# /// Never restart an existing match, just try to remove or spawn cars to match the configuration.
# /// If we are not in the middle of a match, a match will be started. Handy for LAN matches.
    Continue_And_Spawn = 2




from enum import Enum
from queue import Queue, Empty  # TODO(python 3.7+): use SimpleQueue
from threading import Thread
from urllib.parse import ParseResult as URL, urlunparse
import asyncio
import json
import traceback

import websockets
from websockets.client import WebSocketClientProtocol

from rlbot.utils.logging_utils import get_logger
from rlbot.matchcomms.shared import MatchcommsPaths, JSON

class MatchcommsClient:
    def __init__(self, server_root_url: URL):
        self.incoming_broadcast = Queue()
        self.outgoing_broadcast = Queue()

        self.broadcast_url = URL(
            scheme=server_root_url.scheme,
            netloc=server_root_url.netloc,
            path=MatchcommsPaths.BROADCAST,
            params='',
            query='',
            fragment='',
        )

        self.event_loop = asyncio.new_event_loop()
        self.event_loop.create_task(self._run_queue_io())
        self.thread = Thread(target=self.event_loop.run_forever, daemon=True)
        self.thread.start()

    async def _run_queue_io(self):
        try:
            async with websockets.connect(urlunparse(self.broadcast_url)) as websocket:
                io_task = self.event_loop.create_task(asyncio.wait(
                    [
                        read_into_queue(websocket, self.incoming_broadcast),
                        send_from_queue(websocket, self.outgoing_broadcast),
                    ],
                    return_when=asyncio.FIRST_COMPLETED
                ))
                done, pending = await io_task  # should only finish when it is cancelled in close()
                for d in done:
                    d.print_stack()
                assert False, "The io task should only finish when canceled."
        except asyncio.CancelledError as e:
            pass  # We expect a cancel from self.close()

    def close(self):
        # Ensure all tasks are cancelled before stopping.
        # TODO(Python 3.7): always use asyncio.all_tasks
        get_all_tasks = asyncio.all_tasks if hasattr(asyncio, 'all_tasks') else asyncio.Task.all_tasks
        all_tasks = get_all_tasks(loop=self.event_loop)
        for task in all_tasks:
            task.cancel()
        for i in range(100):
            num_canned = sum(task.cancelled() or task.done() for task in all_tasks)
            if num_canned == len(all_tasks):
                break
            self.thread.join(0.01)

        self.event_loop.call_soon_threadsafe(self.event_loop.stop)
        self.thread.join(1)
        assert not self.thread.is_alive()


async def read_into_queue(websocket: WebSocketClientProtocol, incoming: Queue):
    async for message in websocket:
        try:
            incoming.put(json.loads(message))
        except json.decoder.JSONDecodeError:
            print('Failed to decode this message:', repr(message))
            traceback.print_exc()

async def send_from_queue(websocket: WebSocketClientProtocol, outgoing: Queue):
    logger = get_logger('matchcomms_json_encode')
    while True:
        # TODO: use something like https://github.com/aio-libs/janus to avoid polling
        while outgoing.empty():
            await asyncio.sleep(0.01)

        obj = outgoing.get_nowait()
        try:
            json_str = json.dumps(obj) # Serialize the object that was put on the outgoing queue
        except TypeError:
            logger.error(f'This object failed to encode: {obj}\n{traceback.format_exc()}')
        else:
            await websocket.send(json_str)

if __name__ == '__main__':
    from rlbot.matchcomms.server import self_test
    self_test()



from collections import namedtuple
from contextlib import contextmanager
from typing import List

from rlbot.utils.rendering.rendering_manager import RenderingManager

Row = namedtuple('Row', 'exercise_name status_str status_color_func')

@contextmanager
def training_status_renderer_context(exercise_names: List[str], renderman: RenderingManager):
    """
    Ensures that the screen is always cleared, even on fatal errors in code
    that uses this renderer.
    """
    renderer = TrainingStatusRenderer(exercise_names, renderman)
    try:
        yield renderer
    finally:
        renderer.clear_screen()

class TrainingStatusRenderer:
    """
    This class draws the current state of the training exercises
    to the game screen.
    """

    RENDER_GROUP = "TrainingStatusRenderer"
    TEXT_SCALE = 2
    LINE_HEIGHT = 40
    Y_BEGIN = 100
    X_BEGIN = 10
    PAGE_SIZE = 6  # Numer of exercise names drawn at once

    def __init__(self, exercise_names: List[str], renderman: RenderingManager):
        self.renderman = renderman
        self.rows = [Row(name, '', renderman.black) for name in exercise_names]
        self.last_modified_index = 0
        self._render()

    def update(self, index: int, row: Row):
        self.rows[index] = row
        self.last_modified_index = index
        self._render()

    def _render(self):
        if not self.rows:
            self.clear_screen()
            return

        self.renderman.begin_rendering(self.RENDER_GROUP)

        def get_text_y(row_y: int) -> float:
            return self.LINE_HEIGHT * row_y + self.Y_BEGIN

        def draw_string(x_offset: float, row_y: int, text: str, color_func):
            self.renderman.draw_string_2d(
                self.X_BEGIN + x_offset,
                get_text_y(row_y),
                self.TEXT_SCALE,
                self.TEXT_SCALE,
                text,
                color_func()
            )

        if len(self.rows) > self.PAGE_SIZE:
            draw_string(0, -1, f'Progress: {self.last_modified_index+1} / {len(self.rows)}', self.renderman.grey)
        page_begin = (self.last_modified_index // self.PAGE_SIZE) * self.PAGE_SIZE
        for i in range(page_begin, min(page_begin + self.PAGE_SIZE, len(self.rows))):
            row = self.rows[i]
            y = i % self.PAGE_SIZE
            draw_string(0, y, f'[         ] {row.exercise_name}', self.renderman.white)
            draw_string(15, y, row.status_str, row.status_color_func)

        self.renderman.end_rendering()

    def clear_screen(self):
        self.renderman.clear_screen(self.RENDER_GROUP)



import time

from rlbot.botmanager.bot_manager import BotManager


class BotManagerIndependent(BotManager):
    def __init__(self, *args, **kwargs):
        """
        See documentation on BotManager.
        """
        super().__init__(*args, **kwargs)

    def run(self):
        # Get bot module
        agent, agent_class_file = self.load_agent()
        agent.run_independently(self.terminate_request_event)
        if hasattr(agent, 'retire'):
            agent.retire()
        # Give the independent process some time to exit gracefully.
        # We really don't want the setup manager to call kill on us.
        time.sleep(2)
        # If terminated, send callback
        self.termination_complete_event.set()

    def call_agent(self, agent, agent_class):
        pass

    def pull_data_from_game(self):
        pass

    def get_game_time(self):
        pass

    def prepare_for_run(self):
        pass



import hashlib
from pathlib import Path

from rlbot import gateway_util
from rlbot.agents.standalone.standalone_bot_config import StandaloneBotConfig
from rlbot.matchconfig.match_config import PlayerConfig, MatchConfig, MutatorConfig, FLATBUFFER_MAX_INT
from rlbot.parsing.bot_config_bundle import BotConfigBundle
from rlbot.setup_manager import SetupManager


class TestSpawner:
    def __init__(self, python_file: Path, standalone_bot_config: StandaloneBotConfig, bundle: BotConfigBundle):
        self.python_file = python_file
        self.standalone_bot_config = standalone_bot_config
        self.bundle = bundle
        self.player_config: PlayerConfig = None
        self.setup_manager: SetupManager = None
        self.spawn_id = self.create_spawn_id()
        self.player_index = standalone_bot_config.player_index or 0
        self.team = standalone_bot_config.team or 0
        self.name = self.get_bot_name()

    def get_bot_name(self) -> str:
        if self.bundle is not None:
            return self.bundle.name
        if self.standalone_bot_config.name is not None:
            print(
                f'Spawning your bot with the name {self.standalone_bot_config.name} because no config path was provided!')
            return self.standalone_bot_config.name
        print(f'Spawning your bot with the name {self.python_file.name} because no config path was provided!')
        return self.python_file.name

    def create_spawn_id(self):
        """
        We want a spawn id unique to the python file which will be stable across re-runs.
        """
        hash = hashlib.sha1(str(self.python_file).encode('utf-8'))
        number_form = int(hash.hexdigest(), 16)
        return number_form % FLATBUFFER_MAX_INT

    def create_player_config(self, config_path: str) -> PlayerConfig:
        player_config = PlayerConfig()
        player_config.bot = True
        player_config.rlbot_controlled = True
        player_config.bot_skill = 1
        player_config.human_index = 0
        player_config.name = self.name
        player_config.team = 0
        player_config.config_path = config_path
        player_config.spawn_id = self.spawn_id
        return player_config

    def build_match_config(self, player_config_path: str):
        if self.player_config is None:
            self.player_config = self.create_player_config(player_config_path)

        match_config = MatchConfig()
        match_config.player_configs = [self.player_config]
        match_config.game_mode = 'Soccer'
        match_config.game_map = 'DFHStadium'
        match_config.existing_match_behavior = 'Continue And Spawn'
        match_config.mutators = MutatorConfig()
        match_config.enable_state_setting = True
        match_config.enable_rendering = True
        return match_config

    def spawn_bot(self):
        config_path = None
        if self.bundle is not None:
            config_path = self.bundle.config_path

        match_config = self.build_match_config(config_path)

        if self.setup_manager is None:
            self.setup_manager = SetupManager()

            rlbot_gateway_process, _ = gateway_util.find_existing_process()
            if rlbot_gateway_process is None:
                # RLBot.exe is not running yet, we should use the Restart behavior.
                # That avoids a situation where dead cars start piling up when
                # RLBot.exe gets killed and re-launched over and over and lacks context
                # to clean up previous cars.
                match_config.existing_match_behavior = 'Restart'

            self.setup_manager.connect_to_game()

        self.setup_manager.load_match_config(match_config)
        self.setup_manager.start_match()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class QuickChat(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsQuickChat(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = QuickChat()
        x.Init(buf, n + offset)
        return x

    # QuickChat
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # QuickChat
    def QuickChatSelection(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

# /// The index of the player that sent the quick chat
    # QuickChat
    def PlayerIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

# /// True if the chat is team only false if everyone can see it.
    # QuickChat
    def TeamOnly(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # QuickChat
    def MessageIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # QuickChat
    def TimeStamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def QuickChatStart(builder): builder.StartObject(5)
def QuickChatAddQuickChatSelection(builder, quickChatSelection): builder.PrependInt8Slot(0, quickChatSelection, 0)
def QuickChatAddPlayerIndex(builder, playerIndex): builder.PrependInt32Slot(1, playerIndex, 0)
def QuickChatAddTeamOnly(builder, teamOnly): builder.PrependBoolSlot(2, teamOnly, 0)
def QuickChatAddMessageIndex(builder, messageIndex): builder.PrependInt32Slot(3, messageIndex, 0)
def QuickChatAddTimeStamp(builder, timeStamp): builder.PrependFloat32Slot(4, timeStamp, 0.0)
def QuickChatEnd(builder): return builder.EndObject()



from rlbot.utils.structures.utils import create_enum_object

RLBotCoreStatus = create_enum_object(["Success",
                                      "BufferOverfilled",
                                      "MessageLargerThanMax",
                                      "InvalidNumPlayers",
                                      "InvalidBotSkill",
                                      "InvalidHumanIndex",
                                      "InvalidName",
                                      "InvalidTeam",
                                      "InvalidTeamColorID",
                                      "InvalidCustomColorID",
                                      "InvalidGameValues",
                                      "InvalidThrottle",
                                      "InvalidSteer",
                                      "InvalidPitch",
                                      "InvalidYaw",
                                      "InvalidRoll",
                                      "InvalidPlayerIndex",
                                      "InvalidQuickChatPreset",
                                      "InvalidRenderType",
                                      "QuickChatRateExceeded",
                                      "NotInitialized"],
                                     list_name='status_list')



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Expresses the rotation state of an object in Euler angles, with values in radians.
class Rotator(object):
    __slots__ = ['_tab']

    # Rotator
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Rotator
    def Pitch(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Rotator
    def Yaw(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # Rotator
    def Roll(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateRotator(builder, pitch, yaw, roll):
    builder.Prep(4, 12)
    builder.PrependFloat32(roll)
    builder.PrependFloat32(yaw)
    builder.PrependFloat32(pitch)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DropshotTile(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDropshotTile(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DropshotTile()
        x.Init(buf, n + offset)
        return x

    # DropshotTile
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// The amount of damage the tile has sustained.
    # DropshotTile
    def TileState(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

def DropshotTileStart(builder): builder.StartObject(1)
def DropshotTileAddTileState(builder, tileState): builder.PrependInt8Slot(0, tileState, 0)
def DropshotTileEnd(builder): return builder.EndObject()



import random

from rlbot.agents.rlbot_runnable import RLBotRunnable, LOCATIONS_HEADER
from rlbot.messages.flat import MatchSettings
from rlbot.parsing.custom_config import ConfigObject
from rlbot.utils.game_state_util import GameState
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.structures.ball_prediction_struct import BallPrediction
from rlbot.utils.structures.game_data_struct import GameTickPacket
from rlbot.utils.structures.game_interface import GameInterface

SCRIPT_FILE_KEY = "script_file"

class BaseScript(RLBotRunnable):
    """
    A convenience class for building scripts on top of.
    It is NOT required to use this when configuring a script.
    """

    def __init__(self, name):
        super().__init__(name)
        self.logger = get_logger(name)
        self.__key = hash("BaseScript:" + name)
        self.game_tick_packet = GameTickPacket()
        self.ball_prediction = BallPrediction()
        self.game_interface = GameInterface(self.logger)
        self.game_interface.load_interface()
        fake_index = random.randint(100, 10000)  # a number unlikely to conflict with bots or other scripts
        self.renderer = self.game_interface.renderer.get_rendering_manager(bot_index=fake_index, bot_team=2)

    def get_game_tick_packet(self):
        """Gets the latest game tick packet immediately, without blocking."""
        return self.game_interface.update_live_data_packet(self.game_tick_packet)

    def wait_game_tick_packet(self):
        """A blocking call which waits for the next new game tick packet and returns as soon
        as it's available. Will wait for a maximum of 30 milliseconds before giving up and returning
        the packet the framework already has. This is suitable for low-latency update loops."""
        return self.game_interface.fresh_live_data_packet(self.game_tick_packet, 30, self.__key)

    def get_field_info(self):
        """Gets the information about the field.
        This does not change during a match so it only needs to be called once after the everything is loaded."""
        return self.game_interface.get_field_info()

    def set_game_state(self, game_state: GameState):
        self.game_interface.set_game_state(game_state)

    def get_ball_prediction_struct(self) -> BallPrediction:
        """Fetches a prediction of where the ball will go during the next few seconds."""
        return self.game_interface.update_ball_prediction(self.ball_prediction)

    def get_match_settings(self) -> MatchSettings:
        """Gets the current match settings in flatbuffer format. Useful for determining map, game mode,
        mutator settings, etc."""
        return self.game_interface.get_match_settings()

    # Information about @classmethod: https://docs.python.org/3/library/functions.html#classmethod
    @classmethod
    def base_create_agent_configurations(cls) -> ConfigObject:
        """
        This is used when initializing agent config via builder pattern.
        It also calls `create_agent_configurations` that can be used by BaseAgent subclasses for custom configs.
        :return: Returns an instance of a ConfigObject object.
        """

        config = super().base_create_agent_configurations()
        location_config = config.get_header(LOCATIONS_HEADER)

        location_config.add_value(SCRIPT_FILE_KEY, str,
                                  description="Script's python file.")

        cls.create_agent_configurations(config)

        return config



from threading import Thread

from rlbot.messages.flat.FieldInfo import FieldInfo
from rlbot.messages.flat.GameTickPacket import GameTickPacket
from rlbot.messages.flat.MatchSettings import MatchSettings
from rlbot.socket.socket_manager import SocketRelay


class SocketDataReporter:
    def __init__(self):
        self.latest_packet: GameTickPacket = None
        self.latest_field_info: FieldInfo = None
        self.latest_match_settings: MatchSettings = None

        self.socket_relay = SocketRelay()
        self.socket_relay.packet_handlers.append(self._handle_packet)
        self.socket_relay.field_info_handlers.append(self._handle_field_info)
        self.socket_relay.match_settings_handlers.append(self._handle_match_settings)

        self.thread = Thread(target=self.socket_relay.connect_and_run, args=(False, False, False))

        self.thread.start()

    def _handle_packet(self, packet: GameTickPacket):
        self.latest_packet = packet

    def _handle_field_info(self, field_info: FieldInfo):
        self.latest_field_info = field_info

    def _handle_match_settings(self, match_settings: MatchSettings):
        self.latest_match_settings = match_settings

    def disconnect(self):
        self.socket_relay.disconnect()


def get_one_packet() -> GameTickPacket:
    socket_relay = SocketRelay()
    packet = None
    def handle_packet(p):
        nonlocal packet
        packet = p
        socket_relay.disconnect()
    socket_relay.packet_handlers.append(handle_packet)
    socket_relay.connect_and_run(False, False, False)
    return packet



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class BoostPadState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsBoostPadState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BoostPadState()
        x.Init(buf, n + offset)
        return x

    # BoostPadState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// True if the boost can be picked up
    # BoostPadState
    def IsActive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// The number of seconds since the boost has been picked up, or 0.0 if the boost is active.
    # BoostPadState
    def Timer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def BoostPadStateStart(builder): builder.StartObject(2)
def BoostPadStateAddIsActive(builder, isActive): builder.PrependBoolSlot(0, isActive, 0)
def BoostPadStateAddTimer(builder, timer): builder.PrependFloat32Slot(1, timer, 0.0)
def BoostPadStateEnd(builder): return builder.EndObject()



import sys
import logging


DEFAULT_LOGGER = 'rlbot'
logging_level = logging.INFO
FORMAT = "%(asctime)s %(levelname)s:%(name)5s[%(filename)20s:%(lineno)s - %(funcName)20s() ] %(message)s"

logging.getLogger().setLevel(logging.NOTSET)


def get_logger(logger_name, log_creation=True):
    logger = logging.getLogger(logger_name)
    if not logger.handlers:
        ch = logging.StreamHandler()
        ch.setFormatter(logging.Formatter(FORMAT))
        ch.setLevel(logging_level)
        logger.addHandler(ch)
    logging.getLogger().handlers = []
    if logger_name == DEFAULT_LOGGER:
        return logger
    if log_creation:
        logger.debug('creating logger for %s', sys._getframe().f_back)
    return logger


def log_warn(message, args):
    """Logs a warning message using the default logger."""
    get_logger(DEFAULT_LOGGER, log_creation=False).log(logging.WARNING, message, *args)


def log(message):
    get_logger(DEFAULT_LOGGER, log_creation=False).log(logging.INFO, message)



class IncrementingInteger:
    """
    Keeps track of an integer value as it gets incremented.
    """

    def __init__(self, initial_val: int):
        self.value = initial_val

    def increment(self):
        val = self.value
        self.value += 1
        return val



import subprocess
import sys
import warnings
from importlib import import_module
from typing import Set, List

import pkg_resources
import requirements
from pkg_resources import Requirement as PkgRequirement
from requirements.requirement import Requirement


def is_importable(req) -> bool:
    try:
        import_module(req)
        return True
    except:
        return False


# These are not pip packages, but might get specified in a requirements.txt file anyway.
# We'll recommend it because it will help the GUI detect deficiencies in the python environment
# for running that particular bot / script.
SPECIAL_REQUIREMENTS = ['tkinter']


SUPPORTED_SPECIAL_REQUIREMENTS = [req for req in SPECIAL_REQUIREMENTS if is_importable(req)]


def get_installed_packages() -> Set[str]:
    return set([p.project_name for p in pkg_resources.working_set] + SUPPORTED_SPECIAL_REQUIREMENTS)


def get_missing_packages(requirements_file: str = None, special_reqs: List[str] = []) -> List[Requirement]:
    needed = [Requirement.parse_line(sr) for sr in special_reqs]
    if requirements_file is not None:
        with open(requirements_file, 'r') as fd, warnings.catch_warnings():
            warnings.simplefilter("ignore")
            needed += [r for r in requirements.parse(fd) if r.specifier]

    installed = get_installed_packages()
    needed = [r for r in needed if pkg_resources.safe_name(r.name) not in installed]
    return needed


def install_requirements_file(requirements_file):
    return subprocess.call([sys.executable, "-m", "pip", "install", '-r', requirements_file])


def get_packages_needing_upgrade(requirements_file: str) -> List[Requirement]:
    with open(requirements_file, 'r') as fd, warnings.catch_warnings():
        warnings.simplefilter("ignore")
        needed = [r for r in requirements.parse(fd) if r.specifier]

    pkg_resource_map = {p.project_name: p for p in pkg_resources.working_set}
    needs_upgrade = []
    for r in needed:
        safe_name = pkg_resources.safe_name(r.name)
        if safe_name in pkg_resource_map:
            existing = pkg_resource_map[safe_name]
            requirement = PkgRequirement.parse(r.line)
            if existing.version not in requirement.specifier:
                needs_upgrade.append(r)

    return needs_upgrade



import ctypes

from rlbot.utils.structures.bot_input_struct import PlayerInput
from rlbot.utils.structures.game_data_struct import Vector3
from rlbot.utils.structures.start_match_structures import MAX_PLAYERS


class Quaternion(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float),
                ("y", ctypes.c_float),
                ("z", ctypes.c_float),
                ("w", ctypes.c_float)]


class RigidBodyState(ctypes.Structure):
    _fields_ = [("frame", ctypes.c_int),
                ("location", Vector3),
                ("rotation", Quaternion),
                ("velocity", Vector3),
                ("angular_velocity", Vector3)]


class PlayerRigidBodyState(ctypes.Structure):
    _fields_ = [("state", RigidBodyState),
                ("input", PlayerInput)]


class BallRigidBodyState(ctypes.Structure):
    _fields_ = [("state", RigidBodyState)]


class RigidBodyTick(ctypes.Structure):
    _fields_ = [("ball", BallRigidBodyState),
                ("players", PlayerRigidBodyState * MAX_PLAYERS),
                ("num_players", ctypes.c_int)]



import setuptools

__version__ = None  # This will get replaced when reading version.py

exec(open('rlbot/version.py').read())

with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

setuptools.setup(
    name='rlbot',
    packages=setuptools.find_packages(),
    install_requires=[
        'psutil==5.8.0',
        'inputs',
        'websockets',
        'dataclasses',  # Python 3.6 compatibility
        'webdriver_manager',  # For scratch bots
        'selenium',  # For scratch bots
        'requirements-parser',
        'docopt',  # Used by standalone agents
    ],
    version=__version__,
    description='A framework for writing custom Rocket League bots that run offline.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='RLBot Community',
    author_email='rlbotofficial@gmail.com',
    url='https://github.com/RLBot/RLBot',
    keywords=['rocket-league'],
    license='MIT License',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
    ],
    package_data={
        'rlbot': [
            '**/*.json',
            '**/*.png',
            '**/*.md',
            'dll/**/*',
            'dll/*',
            'utils/structures/*.pyi',
            'py.typed',
            'matchconfig/psyonix_presets/*.cfg'
        ]
    },
)



import os

from rlbot.utils.logging_utils import get_logger

logger = get_logger('file util')


def get_python_root() -> str:
    """Gets the path of the python root directory that rlbot lives in."""
    return os.path.dirname(get_rlbot_directory())


def get_rlbot_directory() -> str:
    """Gets the path of the rlbot package directory"""
    return os.path.dirname(os.path.dirname(os.path.realpath(__file__)))


def contains_locked_file(directory: str):
    """
    :return: True if any of the files in the directory are in use. For example, if the dll is injected
    into the game, this will definitely return true.
    """
    for root, subdirs, files in os.walk(directory):
        for filename in files:
            file_path = os.path.join(root, filename)
            try:
                with open(file_path, 'a'):
                    pass
            except IOError:
                logger.debug(f"Locked file: {file_path}")
                return True
    return False



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// A console command which we will try to execute inside Rocket League.
# /// See https://github.com/RLBot/RLBot/wiki/Console-Commands for a list of known commands.
class ConsoleCommand(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsConsoleCommand(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ConsoleCommand()
        x.Init(buf, n + offset)
        return x

    # ConsoleCommand
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ConsoleCommand
    def Command(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return bytes()

def ConsoleCommandStart(builder): builder.StartObject(1)
def ConsoleCommandAddCommand(builder, command): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(command), 0)
def ConsoleCommandEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class CylinderShape(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsCylinderShape(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = CylinderShape()
        x.Init(buf, n + offset)
        return x

    # CylinderShape
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CylinderShape
    def Diameter(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # CylinderShape
    def Height(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def CylinderShapeStart(builder): builder.StartObject(2)
def CylinderShapeAddDiameter(builder, diameter): builder.PrependFloat32Slot(0, diameter, 0.0)
def CylinderShapeAddHeight(builder, height): builder.PrependFloat32Slot(1, height, 0.0)
def CylinderShapeEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class PlayerConfiguration(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayerConfiguration(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PlayerConfiguration()
        x.Init(buf, n + offset)
        return x

    # PlayerConfiguration
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PlayerConfiguration
    def VarietyType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # PlayerConfiguration
    def Variety(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # PlayerConfiguration
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return bytes()

    # PlayerConfiguration
    def Team(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerConfiguration
    def Loadout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .PlayerLoadout import PlayerLoadout
            obj = PlayerLoadout()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// In the case where the requested player index is not available, spawnId will help
# /// the framework figure out what index was actually assigned to this player instead.
    # PlayerConfiguration
    def SpawnId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def PlayerConfigurationStart(builder): builder.StartObject(6)
def PlayerConfigurationAddVarietyType(builder, varietyType): builder.PrependUint8Slot(0, varietyType, 0)
def PlayerConfigurationAddVariety(builder, variety): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(variety), 0)
def PlayerConfigurationAddName(builder, name): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def PlayerConfigurationAddTeam(builder, team): builder.PrependInt32Slot(3, team, 0)
def PlayerConfigurationAddLoadout(builder, loadout): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(loadout), 0)
def PlayerConfigurationAddSpawnId(builder, spawnId): builder.PrependInt32Slot(5, spawnId, 0)
def PlayerConfigurationEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class PredictionSlice(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPredictionSlice(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PredictionSlice()
        x.Init(buf, n + offset)
        return x

    # PredictionSlice
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// The moment in game time that this prediction corresponds to.
# /// This corresponds to 'secondsElapsed' in the GameInfo table.
    # PredictionSlice
    def GameSeconds(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// The predicted location and motion of the object.
    # PredictionSlice
    def Physics(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Physics import Physics
            obj = Physics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def PredictionSliceStart(builder): builder.StartObject(2)
def PredictionSliceAddGameSeconds(builder, gameSeconds): builder.PrependFloat32Slot(0, gameSeconds, 0.0)
def PredictionSliceAddPhysics(builder, physics): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(physics), 0)
def PredictionSliceEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class PlayerClass(object):
    NONE = 0
    RLBotPlayer = 1
    HumanPlayer = 2
    PsyonixBotPlayer = 3
    PartyMemberBotPlayer = 4




import importlib
import inspect
import os
import sys
from pathlib import Path

from rlbot.agents.base_agent import BaseAgent
from rlbot.utils.logging_utils import get_logger


def is_file_under_path(file, path):
    try:
        Path(file).relative_to(path)
        return True
    except ValueError:
        return False


class ExternalClassWrapper:
    """
    Given the absolute path of a python file, this can load the associated module and find a class inside it
    that extends the base class. The module in the target file may assume that its current directory is on
    sys.path and expect import statements to work accordingly.

    Will throw an exception during construction if the module cannot be loaded or the class cannot be located.
    """

    def __init__(self, python_file, base_class):
        self.python_file = python_file
        self.base_class = base_class
        self.logger = get_logger('class_importer')
        self.owned_module_keys = set()
        self._load_external_class_and_record_module_changes()

    def _load_external_class_and_record_module_changes(self):
        keys_before = set(sys.modules.keys())
        self.loaded_class, self.loaded_module = load_external_class(self.python_file, self.base_class)
        # Compute the set of modules which were loaded recursively and had not been seen before
        added_module_keys = set(sys.modules.keys()).difference(keys_before)
        self.owned_module_keys = self.owned_module_keys.union(added_module_keys)

    def get_loaded_class(self):
        """
        Guaranteed to return a valid class that extends base_class.
        """
        return self.loaded_class

    def _reload_owned_modules(self):
        """
        A module is eligible for reload if:
        - It has a python file associated with it that lives under the same directory as the main python_file
        - It's present in self.owned_module_keys, i.e. the main python file appears to have been responsible for its
          insertion into the system module list.

        We do the path check because otherwise we try to reload various low-level modules that have nothing to do with
        the bot logic, and reloading those modules sometimes throws an exception. The owned_module_keys is less
        critical, but it feels safer / faster than doing path checks on the entirety of sys.modules. E.g. what if you
        have two bots in the same folder and only want to reload one of them.
        """
        for module_key in self.owned_module_keys:
            if module_key in sys.modules:
                try:
                    mod = sys.modules[module_key]
                    if hasattr(mod, '__file__') and mod.__file__ and is_file_under_path(mod.__file__, Path(self.python_file).parent):
                        importlib.reload(sys.modules[module_key])
                except Exception as e:
                    self.logger.warning(f"Suppressed error when hot reloading {module_key}: {e}")

    def reload(self):
        dir_name = os.path.dirname(self.python_file)
        sys.path.insert(0, dir_name)
        # Why call this twice? Because the modules would need to be loaded from bottom-up
        # in terms of their dependency tree in order for all changes to manifest in one attempt.
        # That's hard, it's much easier to just repeat it.
        self._reload_owned_modules()
        self._reload_owned_modules()
        del sys.path[0]
        self._load_external_class_and_record_module_changes()


def import_agent(python_file) -> ExternalClassWrapper:
    """
    Imports the first class that extends BaseAgent.

    :param python_file: The absolute path of the bot's main python file
    :return: The agent requested or BaseAgent if there are any problems.
    """
    return import_class_with_base(python_file, BaseAgent)


def import_class_with_base(python_file, base_class) -> ExternalClassWrapper:
    """
    Imports the first class that extends base_class.

    :param python_file: The absolute path of the bot's main python file
    :param base_class: The class that we look for the extension for
    :return: The agent requested or BaseAgent if there are any problems.
    """

    return ExternalClassWrapper(python_file, base_class)


def load_external_class(python_file, base_class):
    """
    Returns a tuple: (subclass of base_class, module)
    """
    loaded_module = load_external_module(python_file)

    # Find a class that extends base_class
    loaded_class = extract_class(loaded_module, base_class)
    return loaded_class, loaded_module


def load_external_module(python_file):
    """
    Returns the loaded module.
    """

    # There's a special case where python_file may be pointing at the base agent definition here in the framework.
    # This is sometimes done as a default and we want to allow it. Short-circuit the logic because
    # loading it as if it's an external class is a real mess.
    if os.path.abspath(python_file) == os.path.abspath(inspect.getfile(BaseAgent)):
        return BaseAgent, BaseAgent.__module__

    if not os.path.isfile(python_file):
        raise FileNotFoundError(f"Could not find file {python_file}!")

    dir_name = os.path.dirname(python_file)
    module_name = os.path.splitext(os.path.basename(python_file))[0]

    # Temporarily modify the sys.path while we load the module so that the module can use import statements naturally
    sys.path.insert(0, dir_name)
    loaded_module = importlib.import_module(module_name)
    del sys.path[0]

    return loaded_module


def extract_class(containing_module, base_class):
    valid_classes = [agent[1] for agent in inspect.getmembers(containing_module, inspect.isclass)
                     if issubclass(agent[1], base_class) and agent[1].__module__ == containing_module.__name__]

    if len(valid_classes) == 0:
        raise ModuleNotFoundError(f"Could not locate a suitable bot class in module {containing_module.__file__}")

    return valid_classes[0]



import os

from rlbot.agents.base_independent_agent import BaseIndependentAgent
from rlbot.botmanager.helper_process_request import HelperProcessRequest


class DroneAgent(BaseIndependentAgent):
    # Path to the hivemind helperprocess python file.
    hive_path = None
    # Bots with the same key will be part of the same hivemind.
    hive_key = None
    # Name of your hivemind that shows up in the console.
    hive_name = None

    def __init__(self, name, team, index):
        super().__init__(name, team, index)
        if self.hive_path is None:
            raise NotImplementedError('You need to specify a path to the hivemind file.')
        if self.hive_key is None:
            raise NotImplementedError('You need to specify a key for your hivemind.')
        if self.hive_name is None:
            raise NotImplementedError('You need to specify a name for your hivemind.')

    def run_independently(self, terminate_request_event):
        pass

    def get_helper_process_request(self) -> HelperProcessRequest:
        if not os.path.isfile(self.hive_path):
            raise FileNotFoundError(f'Could not find file: {self.hive_path}')

        # Appends hive_path to key so that hiveminds in different places don't compete.
        # Appends team to key so that each team has its own hivemind.
        key = f'{self.hive_path}{self.hive_key}{self.team}'

        # Creates request for helper process.
        options = {
            'name': self.hive_name
        }

        return HelperProcessRequest(self.hive_path, key, options=options)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class TileState(object):
    Unknown = 0
# /// The default state of the tiles.
    Filled = 1
# /// The state when a tile has been damaged.
    Damaged = 2
# /// The state of a tile when it is open and a goal can be scored.
    Open = 3




import re
import psutil
from typing import Set, Tuple, Union, List, Dict
import platform
from rlbot.utils.logging_utils import get_logger, DEFAULT_LOGGER
from rlbot.botmanager.agent_metadata import AgentMetadata


def configure_processes(agent_metadata_map: Dict[int, AgentMetadata], logger):
    """
    This will update the priority and CPU affinity of the processes owned by bots to try to achieve fairness and
    good performance.

    :param agent_metadata_map: A mapping of player index to agent metadata, including a list of owned process ids.
    """

    team_pids_map = {}

    for player_index, data in agent_metadata_map.items():
        team = data.team
        if team not in team_pids_map:
            team_pids_map[team] = set()
        team_pids_map[team].update(data.pids)

    shared_pids = set()

    if len(team_pids_map) >= 2 and len(get_team_cpus(0)) > 0:
        # Sort into three sets of pids: team 0 exclusives, team 1 exclusives, and shared pids
        # All pids will be assigned high priority
        # Team exclusive pids will be bound to a subset of cpus so they can't adversely affect the opposite team.

        for team, team_set in team_pids_map.items():
            if not shared_pids:
                shared_pids.update(team_set)
            else:
                shared_pids.intersection_update(team_set)

        for team, team_set in team_pids_map.items():
            team_set -= shared_pids

        for team, team_pids in team_pids_map.items():
            team_cpus = get_team_cpus(team)
            for pid in team_pids:
                try:
                    p = psutil.Process(pid)
                    configure_process(p, team_cpus)
                except Exception as e:
                    get_logger(DEFAULT_LOGGER).info(e)
    else:
        # Consider everything a shared pid, because we are not in a position to split up cpus.
        for team, team_set in team_pids_map.items():
            shared_pids.update(team_set)

    for pid in shared_pids:
        try:
            p = psutil.Process(pid)  # Allow the process to run at high priority
            if platform.system() == 'Windows':
                p.nice(psutil.HIGH_PRIORITY_CLASS)
        except Exception as e:
            get_logger(DEFAULT_LOGGER).info(e)


def configure_process(proc: psutil.Process, cores: List[int], infer_multi_team=False):
    try:
        if infer_multi_team:
            existing_affinity = proc.cpu_affinity()
            if len(existing_affinity) < psutil.cpu_count():
                # The process in question has already been pinned to a subset of the CPUs.
                # This might be a process spanning multiple teams, so we will set affinity to
                # the combination of existing and newly requested cores.
                cores = sorted(set(existing_affinity + cores))
        proc.cpu_affinity(cores)

        if platform.system() == 'Windows':
            proc.nice(psutil.HIGH_PRIORITY_CLASS)  # Allow the process to run at high priority
    except Exception as e:
        get_logger(DEFAULT_LOGGER).info(e)


def get_team_cpus(team: int) -> List[int]:
    cpu_count = psutil.cpu_count()
    cpus_per_team = cpu_count // 3
    team_cpu_offset = cpus_per_team * team
    return list(range(cpu_count - cpus_per_team - team_cpu_offset, cpu_count - team_cpu_offset))


def extract_all_pids(agent_metadata_map):
    pids = set()
    for player_index, data in agent_metadata_map.items():
        pids.update(data.pids)
    return pids


def is_process_running(program, scriptname, required_args: Set[str]) -> Tuple[bool, Union[psutil.Process, None]]:
    # Find processes which contain the program or script name.
    matching_processes = []
    for process in psutil.process_iter():
        try:
            p = process.name()
            if program in p or scriptname in p:
                matching_processes.append(process)
        except psutil.NoSuchProcess:
            continue
    # If matching processes were found, check for correct arguments.
    if len(matching_processes) != 0:
        for process in matching_processes:
            try:
                args = process.cmdline()[1:]
                for required_arg in required_args:
                    matching_args = [arg for arg in args if re.match(required_arg, arg, flags=re.IGNORECASE) is not None]
                    # Skip this process because it does not have a matching required argument.
                    if len(matching_args) == 0:
                        break
                else:
                    # If this process has not been skipped, it matches all arguments.
                    return True, process
            except psutil.AccessDenied:
                print(f"Access denied when trying to look at cmdline of {process}!")
        # If we didn't return yet it means all matching programs were skipped.
        raise WrongProcessArgs(f"{program} is not running with required arguments: {required_args}!")
    # No matching processes.
    return False, None


class WrongProcessArgs(UserWarning):
    pass



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class BoostOption(object):
    Normal_Boost = 0
    Unlimited_Boost = 1
    Slow_Recharge = 2
    Rapid_Recharge = 3
    No_Boost = 4




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class Vector3Partial(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsVector3Partial(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Vector3Partial()
        x.Init(buf, n + offset)
        return x

    # Vector3Partial
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Vector3Partial
    def X(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Vector3Partial
    def Y(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Vector3Partial
    def Z(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def Vector3PartialStart(builder): builder.StartObject(3)
def Vector3PartialAddX(builder, x): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(x), 0)
def Vector3PartialAddY(builder, y): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(y), 0)
def Vector3PartialAddZ(builder, z): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(z), 0)
def Vector3PartialEnd(builder): return builder.EndObject()



from dataclasses import dataclass

from flatbuffers import Builder
from rlbot.messages.flat import LoadoutPaint, PlayerLoadout, Color as FlatColor
from rlbot.utils.structures.start_match_structures import PlayerConfiguration
from rlbot.utils.structures.start_match_structures import Color as ColorStruct


class LoadoutConfig:
    def __init__(self):
        self.team_color_id: int = 0
        self.custom_color_id: int = 0
        self.car_id: int = 0
        self.decal_id: int = 0
        self.wheels_id: int = 0
        self.boost_id: int = 0
        self.antenna_id: int = 0
        self.hat_id: int = 0
        self.paint_finish_id: int = 0
        self.custom_finish_id: int = 0
        self.engine_audio_id: int = 0
        self.trails_id: int = 0
        self.goal_explosion_id: int = 0
        self.paint_config: LoadoutPaintConfig = LoadoutPaintConfig()
        self.primary_color_lookup: Color = None
        self.secondary_color_lookup: Color = None

    def write(self, player_configuration: PlayerConfiguration):
        player_configuration.team_color_id = self.team_color_id
        player_configuration.custom_color_id = self.custom_color_id
        player_configuration.car_id = self.car_id
        player_configuration.decal_id = self.decal_id
        player_configuration.wheels_id = self.wheels_id
        player_configuration.boost_id = self.boost_id
        player_configuration.antenna_id = self.antenna_id
        player_configuration.hat_id = self.hat_id
        player_configuration.paint_finish_id = self.paint_finish_id
        player_configuration.custom_finish_id = self.custom_finish_id
        player_configuration.engine_audio_id = self.engine_audio_id
        player_configuration.trails_id = self.trails_id
        player_configuration.goal_explosion_id = self.goal_explosion_id

        if self.paint_config:
            self.paint_config.write(player_configuration)

        if self.primary_color_lookup:
            player_configuration.use_rgb_lookup = True
            self.primary_color_lookup.write(player_configuration.primary_color_lookup)

        if self.secondary_color_lookup:
            player_configuration.use_rgb_lookup = True
            self.secondary_color_lookup.write(player_configuration.secondary_color_lookup)

    def write_to_flatbuffer(self, builder: Builder):
        if self.paint_config:
            paint_offset = self.paint_config.write_to_flatbuffer(builder)

        if self.primary_color_lookup:
            primary_color_offset = self.primary_color_lookup.write_to_flatbuffer(builder)
        if self.secondary_color_lookup:
            secondary_color_offset = self.secondary_color_lookup.write_to_flatbuffer(builder)

        PlayerLoadout.PlayerLoadoutStart(builder)
        PlayerLoadout.PlayerLoadoutAddTeamColorId(builder, self.team_color_id)
        PlayerLoadout.PlayerLoadoutAddCustomColorId(builder, self.custom_color_id)
        PlayerLoadout.PlayerLoadoutAddCarId(builder, self.car_id)
        PlayerLoadout.PlayerLoadoutAddDecalId(builder, self.decal_id)
        PlayerLoadout.PlayerLoadoutAddWheelsId(builder, self.wheels_id)
        PlayerLoadout.PlayerLoadoutAddBoostId(builder, self.boost_id)
        PlayerLoadout.PlayerLoadoutAddAntennaId(builder, self.antenna_id)
        PlayerLoadout.PlayerLoadoutAddHatId(builder, self.hat_id)
        PlayerLoadout.PlayerLoadoutAddPaintFinishId(builder, self.paint_finish_id)
        PlayerLoadout.PlayerLoadoutAddCustomFinishId(builder, self.custom_finish_id)
        PlayerLoadout.PlayerLoadoutAddEngineAudioId(builder, self.engine_audio_id)
        PlayerLoadout.PlayerLoadoutAddTrailsId(builder, self.trails_id)
        PlayerLoadout.PlayerLoadoutAddGoalExplosionId(builder, self.goal_explosion_id)
        if self.paint_config:
            PlayerLoadout.PlayerLoadoutAddLoadoutPaint(builder, paint_offset)
        if self.primary_color_lookup:
            PlayerLoadout.PlayerLoadoutAddPrimaryColorLookup(builder, primary_color_offset)
        if self.secondary_color_lookup:
            PlayerLoadout.PlayerLoadoutAddSecondaryColorLookup(builder, secondary_color_offset)
        return PlayerLoadout.PlayerLoadoutEnd(builder)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

class LoadoutPaintConfig:
    def __init__(self):
        self.car_paint_id: int = 0
        self.decal_paint_id: int = 0
        self.wheels_paint_id: int = 0
        self.boost_paint_id: int = 0
        self.antenna_paint_id: int = 0
        self.hat_paint_id: int = 0
        self.trails_paint_id: int = 0
        self.goal_explosion_paint_id: int = 0

    def write(self, player_configuration: PlayerConfiguration):
        player_configuration.car_paint_id = self.car_paint_id
        player_configuration.decal_paint_id = self.decal_paint_id
        player_configuration.wheels_paint_id = self.wheels_paint_id
        player_configuration.boost_paint_id = self.boost_paint_id
        player_configuration.antenna_paint_id = self.antenna_paint_id
        player_configuration.hat_paint_id = self.hat_paint_id
        player_configuration.trails_paint_id = self.trails_paint_id
        player_configuration.goal_explosion_paint_id = self.goal_explosion_paint_id

    def write_to_flatbuffer(self, builder: Builder):
        paint = self
        LoadoutPaint.LoadoutPaintStart(builder)
        if paint:
            LoadoutPaint.LoadoutPaintAddCarPaintId(builder, paint.car_paint_id)
            LoadoutPaint.LoadoutPaintAddDecalPaintId(builder, paint.decal_paint_id)
            LoadoutPaint.LoadoutPaintAddWheelsPaintId(builder, paint.wheels_paint_id)
            LoadoutPaint.LoadoutPaintAddBoostPaintId(builder, paint.boost_paint_id)
            LoadoutPaint.LoadoutPaintAddAntennaPaintId(builder, paint.antenna_paint_id)
            LoadoutPaint.LoadoutPaintAddHatPaintId(builder, paint.hat_paint_id)
            LoadoutPaint.LoadoutPaintAddTrailsPaintId(builder, paint.trails_paint_id)
            LoadoutPaint.LoadoutPaintAddGoalExplosionPaintId(builder, paint.goal_explosion_paint_id)
        return LoadoutPaint.LoadoutPaintEnd(builder)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


@dataclass
class Color:
    red: int
    green: int
    blue: int
    alpha: int

    def write(self, color: ColorStruct):
        color.r = self.red
        color.g = self.green
        color.b = self.blue
        color.a = self.alpha

    def write_to_flatbuffer(self, builder: Builder):
        FlatColor.ColorStart(builder)
        FlatColor.ColorAddR(builder, self.red)
        FlatColor.ColorAddG(builder, self.green)
        FlatColor.ColorAddB(builder, self.blue)
        FlatColor.ColorAddA(builder, self.alpha)
        return FlatColor.ColorEnd(builder)






# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DesiredCarState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDesiredCarState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DesiredCarState()
        x.Init(buf, n + offset)
        return x

    # DesiredCarState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DesiredCarState
    def Physics(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .DesiredPhysics import DesiredPhysics
            obj = DesiredPhysics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredCarState
    def BoostAmount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredCarState
    def Jumped(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Bool import Bool
            obj = Bool()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredCarState
    def DoubleJumped(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Bool import Bool
            obj = Bool()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def DesiredCarStateStart(builder): builder.StartObject(4)
def DesiredCarStateAddPhysics(builder, physics): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(physics), 0)
def DesiredCarStateAddBoostAmount(builder, boostAmount): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(boostAmount), 0)
def DesiredCarStateAddJumped(builder, jumped): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(jumped), 0)
def DesiredCarStateAddDoubleJumped(builder, doubleJumped): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(doubleJumped), 0)
def DesiredCarStateEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class QuickChatSelection(object):
    Information_IGotIt = 0
    Information_NeedBoost = 1
    Information_TakeTheShot = 2
    Information_Defending = 3
    Information_GoForIt = 4
    Information_Centering = 5
    Information_AllYours = 6
    Information_InPosition = 7
    Information_Incoming = 8
    Compliments_NiceShot = 9
    Compliments_GreatPass = 10
    Compliments_Thanks = 11
    Compliments_WhatASave = 12
    Compliments_NiceOne = 13
    Compliments_WhatAPlay = 14
    Compliments_GreatClear = 15
    Compliments_NiceBlock = 16
    Reactions_OMG = 17
    Reactions_Noooo = 18
    Reactions_Wow = 19
    Reactions_CloseOne = 20
    Reactions_NoWay = 21
    Reactions_HolyCow = 22
    Reactions_Whew = 23
    Reactions_Siiiick = 24
    Reactions_Calculated = 25
    Reactions_Savage = 26
    Reactions_Okay = 27
    Apologies_Cursing = 28
    Apologies_NoProblem = 29
    Apologies_Whoops = 30
    Apologies_Sorry = 31
    Apologies_MyBad = 32
    Apologies_Oops = 33
    Apologies_MyFault = 34
    PostGame_Gg = 35
    PostGame_WellPlayed = 36
    PostGame_ThatWasFun = 37
    PostGame_Rematch = 38
    PostGame_OneMoreGame = 39
    PostGame_WhatAGame = 40
    PostGame_NiceMoves = 41
    PostGame_EverybodyDance = 42
# /// Custom text chats made by bot makers
    MaxPysonixQuickChatPresets = 43
# /// Waste of CPU cycles
    Custom_Toxic_WasteCPU = 44
# /// Git gud*
    Custom_Toxic_GitGut = 45
# /// De-Allocate Yourself
    Custom_Toxic_DeAlloc = 46
# /// 404: Your skill not found
    Custom_Toxic_404NoSkill = 47
# /// Get a virus
    Custom_Toxic_CatchVirus = 48
# /// Passing!
    Custom_Useful_Passing = 49
# /// Faking!
    Custom_Useful_Faking = 50
# /// Demoing!
    Custom_Useful_Demoing = 51
# /// BOOPING
    Custom_Useful_Bumping = 52
# /// The chances of that was 47525 to 1*
    Custom_Compliments_TinyChances = 53
# /// Who upped your skill level?
    Custom_Compliments_SkillLevel = 54
# /// Your programmer should be proud
    Custom_Compliments_proud = 55
# /// You're the GC of Bots
    Custom_Compliments_GC = 56
# /// Are you <Insert Pro>Bot? *
    Custom_Compliments_Pro = 57
# /// Lag
    Custom_Excuses_Lag = 58
# /// Ghost inputs
    Custom_Excuses_GhostInputs = 59
# /// RIGGED
    Custom_Excuses_Rigged = 60
# /// Mafia plays!
    Custom_Toxic_MafiaPlays = 61
# /// Yeet!
    Custom_Exclamation_Yeet = 62




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class BallInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsBallInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BallInfo()
        x.Init(buf, n + offset)
        return x

    # BallInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # BallInfo
    def Physics(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Physics import Physics
            obj = Physics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # BallInfo
    def LatestTouch(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Touch import Touch
            obj = Touch()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # BallInfo
    def DropShotInfo(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .DropShotBallInfo import DropShotBallInfo
            obj = DropShotBallInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # BallInfo
    def ShapeType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # BallInfo
    def Shape(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

def BallInfoStart(builder): builder.StartObject(5)
def BallInfoAddPhysics(builder, physics): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(physics), 0)
def BallInfoAddLatestTouch(builder, latestTouch): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(latestTouch), 0)
def BallInfoAddDropShotInfo(builder, dropShotInfo): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(dropShotInfo), 0)
def BallInfoAddShapeType(builder, shapeType): builder.PrependUint8Slot(3, shapeType, 0)
def BallInfoAddShape(builder, shape): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(shape), 0)
def BallInfoEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class MatchSettings(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMatchSettings(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MatchSettings()
        x.Init(buf, n + offset)
        return x

    # MatchSettings
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # MatchSettings
    def PlayerConfigurations(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .PlayerConfiguration import PlayerConfiguration
            obj = PlayerConfiguration()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # MatchSettings
    def PlayerConfigurationsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MatchSettings
    def GameMode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def GameMap(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def SkipReplays(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def InstantStart(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def MutatorSettings(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .MutatorSettings import MutatorSettings
            obj = MutatorSettings()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # MatchSettings
    def ExistingMatchBehavior(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def EnableLockstep(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def EnableRendering(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def EnableStateSetting(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # MatchSettings
    def AutoSaveReplay(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// The name of a upk file, like UtopiaStadium_P, which should be loaded.
# /// If specified, this overrides gameMap. On Steam version of Rocket League,
# /// this can be used to load custom map files, but on Epic version it only
# /// works on the Psyonix maps. Still useful because maintaining the gameMap
# /// enum as new Psyonix maps are added is annoying.
    # MatchSettings
    def GameMapUpk(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return bytes()

def MatchSettingsStart(builder): builder.StartObject(12)
def MatchSettingsAddPlayerConfigurations(builder, playerConfigurations): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(playerConfigurations), 0)
def MatchSettingsStartPlayerConfigurationsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def MatchSettingsAddGameMode(builder, gameMode): builder.PrependInt8Slot(1, gameMode, 0)
def MatchSettingsAddGameMap(builder, gameMap): builder.PrependInt8Slot(2, gameMap, 0)
def MatchSettingsAddSkipReplays(builder, skipReplays): builder.PrependBoolSlot(3, skipReplays, 0)
def MatchSettingsAddInstantStart(builder, instantStart): builder.PrependBoolSlot(4, instantStart, 0)
def MatchSettingsAddMutatorSettings(builder, mutatorSettings): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(mutatorSettings), 0)
def MatchSettingsAddExistingMatchBehavior(builder, existingMatchBehavior): builder.PrependInt8Slot(6, existingMatchBehavior, 0)
def MatchSettingsAddEnableLockstep(builder, enableLockstep): builder.PrependBoolSlot(7, enableLockstep, 0)
def MatchSettingsAddEnableRendering(builder, enableRendering): builder.PrependBoolSlot(8, enableRendering, 0)
def MatchSettingsAddEnableStateSetting(builder, enableStateSetting): builder.PrependBoolSlot(9, enableStateSetting, 0)
def MatchSettingsAddAutoSaveReplay(builder, autoSaveReplay): builder.PrependBoolSlot(10, autoSaveReplay, 0)
def MatchSettingsAddGameMapUpk(builder, gameMapUpk): builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(gameMapUpk), 0)
def MatchSettingsEnd(builder): return builder.EndObject()



from rlbot.parsing.custom_config import ConfigObject

LOCATIONS_HEADER = 'Locations'
DETAILS_HEADER = 'Details'
PYTHON_FILE_KEY = 'python_file'
REQUIREMENTS_FILE_KEY = 'requirements_file'
LOGO_FILE_KEY = 'logo_file'
NAME_KEY = "name"
SUPPORTS_EARLY_START_KEY = "supports_early_start"
REQUIRES_TKINTER = "requires_tkinter"
USE_VIRTUAL_ENVIRONMENT_KEY = "use_virtual_environment"


class RLBotRunnable:
    name = None

    def __init__(self, name):
        self.name = name

    def init_match_config(self, match_config: 'MatchConfig'):
        """
        Override this method if you would like to be informed of what config was used to start the match.
        Useful for knowing what map you're on, mutators, etc.
        """
        pass

    def initialize_agent(self):
        """
        Called for all heaver initialization that needs to happen.
        The config is fully loaded at this point
        """
        pass

    def retire(self):
        """Called after the game ends"""
        pass

    @staticmethod
    def create_agent_configurations(config: ConfigObject):
        """
        If your bot needs to add custom configurations, you may override this and use the `config` object.
        :param config: A ConfigObject instance.
        """
        pass

    def is_hot_reload_enabled(self):
        """
        If true, the framework will watch all your python files for modifications and instantly reload your bot
        so that the logic changes take effect. You may wish to disable this if you're concerned about performance.
        """
        return True

    # Information about @classmethod: https://docs.python.org/3/library/functions.html#classmethod
    @classmethod
    def base_create_agent_configurations(cls) -> ConfigObject:
        """
        This is used when initializing agent config via builder pattern.
        It also calls `create_agent_configurations` that can be used by BaseAgent subclasses for custom configs.
        :return: Returns an instance of a ConfigObject object.
        """
        config = ConfigObject()
        location_config = config.add_header_name(LOCATIONS_HEADER)
        location_config.add_value(PYTHON_FILE_KEY, str,
                                  description="Bot's python file.\nOnly need this if RLBot controlled")
        location_config.add_value(REQUIREMENTS_FILE_KEY, str,
                                  description="Python requirements.txt file listing needed dependencies.")
        location_config.add_value(NAME_KEY, str, default='nameless',
                                  description='The name that will be displayed in game')
        location_config.add_value(LOGO_FILE_KEY, str,
                                  description="Location of an image file to use as your bot's logo")
        location_config.add_value(SUPPORTS_EARLY_START_KEY, bool,
                                  description="True if this bot can be started before the Rocket League match begins.")
        location_config.add_value(REQUIRES_TKINTER, bool,
                                  description="True if the tkinter library is needed.")
        location_config.add_value(USE_VIRTUAL_ENVIRONMENT_KEY, bool,
                                  description="True if the runnable wants to run in a virtual environment.")

        details_config = config.add_header_name(DETAILS_HEADER)
        details_config.add_value('developer', str, description="Name of the bot's creator/developer")
        details_config.add_value('description', str, description="Short description of the bot")
        details_config.add_value('fun_fact', str, description="Fun fact about the bot")
        details_config.add_value('github', str, description="Link to github repository")
        details_config.add_value('language', str, description="Programming language")

        cls.create_agent_configurations(config)

        return config



from rlbot.agents.base_agent import BaseAgent


class BaseIndependentAgent(BaseAgent):
    """
    Extend this class if you want to manage calling the dll yourself.
    """

    def run_independently(self, terminate_request_event):
        raise NotImplementedError

    def is_hot_reload_enabled(self):
        return False



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// A psyonix bot, e.g. All Star bot
class PsyonixBotPlayer(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPsyonixBotPlayer(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PsyonixBotPlayer()
        x.Init(buf, n + offset)
        return x

    # PsyonixBotPlayer
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PsyonixBotPlayer
    def BotSkill(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def PsyonixBotPlayerStart(builder): builder.StartObject(1)
def PsyonixBotPlayerAddBotSkill(builder, botSkill): builder.PrependFloat32Slot(0, botSkill, 0.0)
def PsyonixBotPlayerEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class TeamInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTeamInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TeamInfo()
        x.Init(buf, n + offset)
        return x

    # TeamInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TeamInfo
    def TeamIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

# /// number of goals scored.
    # TeamInfo
    def Score(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def TeamInfoStart(builder): builder.StartObject(2)
def TeamInfoAddTeamIndex(builder, teamIndex): builder.PrependInt32Slot(0, teamIndex, 0)
def TeamInfoAddScore(builder, score): builder.PrependInt32Slot(1, score, 0)
def TeamInfoEnd(builder): return builder.EndObject()
